/***************************************************************************
                          Calc.cpp  -  description
                             -------------------
    begin                : Wed Nov 15 2000
    copyright            : (C) 2000 by Gyorgy Molnar
    email                : lindev@home.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "Calc.h"
#include "CalcInternal.h"
#include "Misc.h"
#include <gtk/gtk.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <stdlib.h>
#include <errno.h>

#define STACK_DOUBLE_SIZE 15
#define STACK_INT_SIZE    15

class CStackIntExp
{
  StackIntItem        m_Stack[STACK_INT_SIZE];
  int                 m_iStackLevel;
  int                 m_Nest[STACK_INT_SIZE];
  int                 m_iNestLevel;
  int GetPriority(KeyTokens tcOp);
public:
  CStackIntExp();
  ~CStackIntExp();
  void Delete(void);
  int AddOp(KeyTokens tcOp, int & dRes);
  int AddOp(int dOp, int & dRes);
  int ExeAll(int & dRes);
  int Exe(int dArg1, int dArg2, KeyTokens tcOp, int & dRes);
};

CStackIntExp::CStackIntExp()
{
  Delete();
}

CStackIntExp::~CStackIntExp()
{
}

void CStackIntExp::Delete()
{
  m_iStackLevel = 0;
  m_iNestLevel = 0;
}

int CStackIntExp::AddOp(int iOp, int & iRes)
{
  int iRet = 0;
  if(m_iStackLevel && (m_iStackLevel >= STACK_INT_SIZE ||
                        m_Stack[m_iStackLevel - 1].tcOp == tcNull))
    iRet = -1;
  else
  {
    m_Stack[m_iStackLevel].iNumber = iOp;
    m_Stack[m_iStackLevel].tcOp    = tcNull;
    m_iStackLevel++;
    iRes = iOp;
  }
  return iRet;
}

int CStackIntExp::GetPriority(KeyTokens tcOp)
{
  int i = 0;
  while((PryListInt[i].tcOp != tcOp) && (PryListInt[i].tcOp != tcNull))
    i++;
  return PryListInt[i].iPryLevel;
}

int CStackIntExp::AddOp(KeyTokens tcOp, int & iRes)
{
  int i, iRet = -1;
  int iLowLimit = m_iNestLevel ? m_Nest[m_iNestLevel - 1] : 1;
  for(i = m_iStackLevel; iLowLimit < i; i--)
  {
    int iRes;
    if(m_Stack[i - 1].tcOp != tcNull)
      return 0;
    if(GetPriority(m_Stack[i - 2].tcOp) < GetPriority(tcOp))
      break;
    if(Exe(m_Stack[i - 2].iNumber,
           m_Stack[i - 1].iNumber,
           m_Stack[i - 2].tcOp, iRes) < 0)
      return 0;
    m_Stack[i - 2].iNumber = iRes;
    m_Stack[i - 2].tcOp    = tcNull;
  }
  m_iStackLevel = i;
  if(m_iStackLevel && m_Stack[m_iStackLevel - 1].tcOp == tcNull)
  {
    m_Stack[m_iStackLevel - 1].tcOp = tcOp;
    iRes = m_Stack[m_iStackLevel - 1].iNumber;
    iRet = 0;
  }
  return iRet;
}

int CStackIntExp::ExeAll(int & iRes)
{
  int i, iRet;
  for(i = m_iStackLevel; i > 1; i--)
  {
    int iRes;
    if(m_Stack[i - 1].tcOp != tcNull)
      return -1;
    if(Exe(m_Stack[i - 2].iNumber,
           m_Stack[i - 1].iNumber,
           m_Stack[i - 2].tcOp, iRes) < 0)
      return -1;
    m_Stack[i - 2].iNumber = iRes;
    m_Stack[i - 2].tcOp    = tcNull;
  }
  if(i == 0)
    iRes = 0;
  else
    iRes = m_Stack[0].iNumber;
  Delete();
  iRet = 0;
  return iRet;
}

int CStackIntExp::Exe(int iArg1, int iArg2, KeyTokens tcOp, int & iRes)
{
  int iRet = 0;
  switch(tcOp)
  {
  case tcAdd:
    iRes = (unsigned int)iArg1 + (unsigned int)iArg2;
    break;
  case tcSub:
    iRes = (unsigned int)iArg1 - (unsigned int)iArg2;
    break;
  case tcMul:
    iRes = (unsigned int)iArg1 * (unsigned int)iArg2;
    break;
  case tcDiv:
    if(iArg2 != 0)
      iRes = (unsigned int)iArg1 / (unsigned int)iArg2;
    else
      iRes = 0;
    break;
  case tcMod:
    if(iArg2 != 0)
      iRes = (unsigned int)iArg1 % (unsigned int)iArg2;
    else
      iRes = 0;
    break;
  case tcAnd:
    iRes = (unsigned int)iArg1 & (unsigned int)iArg2;
    break;
  case tcOr:
    iRes = (unsigned int)iArg1 | (unsigned int)iArg2;
    break;
  case tcXor:
    iRes = (unsigned int)iArg1 ^ (unsigned int)iArg2;
    break;
  default:
    iRet = -1;
   }
  return iRet;
}

//**************************************************************************************************


class CStackDoubleExp
{
  StackDoubleItem     m_Stack[STACK_DOUBLE_SIZE];
  int                 m_iStackLevel;
  int                 m_Nest[STACK_DOUBLE_SIZE];
  int                 m_iNestLevel;
  int GetPriority(KeyTokens tcOp);
public:
  CStackDoubleExp();
  ~CStackDoubleExp();
  void Delete(void);
  int AddOp(KeyTokens tcOp, double & dRes);
  int AddOp(double dOp, double & dRes);
  int AddLP(double & dRes);
  int AddRP(double & dRes);
  int ExeAll(double & dRes);
  int Exe(double dArg1, double dArg2, KeyTokens tcOp, double & dRes);
};

CStackDoubleExp::CStackDoubleExp()
{
  Delete();
}

CStackDoubleExp::~CStackDoubleExp()
{
}

void CStackDoubleExp::Delete()
{
  m_iStackLevel = 0;
  m_iNestLevel = 0;
}

int CStackDoubleExp::AddOp(double dOp, double & dRes)
{
  int iRet = 0;
  if(m_iStackLevel && (m_iStackLevel >= STACK_DOUBLE_SIZE ||
                        m_Stack[m_iStackLevel - 1].tcOp == tcNull))
    iRet = -1;
  else
  {
    m_Stack[m_iStackLevel].dNumber = dOp;
    m_Stack[m_iStackLevel].tcOp = tcNull;
    m_iStackLevel++;
    dRes = dOp;
  }
  return iRet;
}

int CStackDoubleExp::GetPriority(KeyTokens tcOp)
{
  int i = 0;
  while((PryList[i].tcOp != tcOp) && (PryList[i].tcOp != tcNull))
    i++;
  return PryList[i].iPryLevel;
}

int CStackDoubleExp::AddLP(double & dRes)
{
  if(m_iNestLevel >= STACK_DOUBLE_SIZE)
    return -1;
  if(!m_iStackLevel || m_Stack[m_iStackLevel - 1].tcOp != tcNull)
  {
    m_Nest[m_iNestLevel++] = m_iStackLevel + 1;
    dRes = 0;
    return 0;
  }
  return -1;
}

int CStackDoubleExp::AddRP(double & dRes)
{
  int i, iLowLimit = m_iNestLevel ? m_Nest[m_iNestLevel - 1] : 1;
  if(AddOp(dRes, dRes) < 0)
    return -1;
  for(i = m_iStackLevel; iLowLimit < i; i--)
  {
    if(m_Stack[i - 1].tcOp != tcNull)
      return 0;
    if(Exe(m_Stack[i - 2].dNumber,
           m_Stack[i - 1].dNumber,
           m_Stack[i - 2].tcOp, dRes) < 0)
      return 0;
    m_Stack[i - 2].dNumber = dRes;
    m_Stack[i - 2].tcOp    = tcNull;
  }
  m_iStackLevel = i - 1;
  if(m_iNestLevel) m_iNestLevel--;
  return 1;
}

int CStackDoubleExp::AddOp(KeyTokens tcOp, double & dRes)
{
  int i, iRet = -1;
  int iLowLimit = m_iNestLevel ? m_Nest[m_iNestLevel - 1] : 1;
  for(i = m_iStackLevel; iLowLimit < i; i--)
  {
    double dRes;
    if(m_Stack[i - 1].tcOp != tcNull)
      return 0;
    if(GetPriority(m_Stack[i - 2].tcOp) < GetPriority(tcOp))
      break;
    if(Exe(m_Stack[i - 2].dNumber,
           m_Stack[i - 1].dNumber,
           m_Stack[i - 2].tcOp, dRes) < 0)
      return 0;
    m_Stack[i - 2].dNumber = dRes;
    m_Stack[i - 2].tcOp    = tcNull;
  }
  m_iStackLevel = i;
  if(m_iStackLevel && m_Stack[m_iStackLevel - 1].tcOp == tcNull)
  {
    m_Stack[m_iStackLevel - 1].tcOp = tcOp;
    dRes = m_Stack[m_iStackLevel - 1].dNumber;
    iRet = 0;
  }
  return iRet;
}

int CStackDoubleExp::ExeAll(double & dRes)
{
  int i, iRet;
  for(i = m_iStackLevel; i > 1; i--)
  {
    double dRes;
    if(m_Stack[i - 1].tcOp != tcNull)
      return -1;
    if(Exe(m_Stack[i - 2].dNumber,
           m_Stack[i - 1].dNumber,
           m_Stack[i - 2].tcOp, dRes) < 0)
      return -1;
    m_Stack[i - 2].dNumber = dRes;
    m_Stack[i - 2].tcOp    = tcNull;
  }
  if(i == 0)
    dRes = 0;
  else
    dRes = m_Stack[0].dNumber;
  Delete();
  iRet = 0;
  return iRet;
}

int CStackDoubleExp::Exe(double dArg1, double dArg2, KeyTokens tcOp, double & dRes)
{
  int iRet = 0;
  switch(tcOp)
  {
  case tcAdd:
    dRes = dArg1 + dArg2;
    break;
  case tcSub:
    dRes = dArg1 - dArg2;
    break;
  case tcMul:
    dRes = dArg1 * dArg2;
    break;
  case tcDiv:
    if(dArg2 != 0)
      dRes = dArg1 / dArg2;
    else
      dRes = 0;
    break;
  case tcMod:
    dRes = fmod(dArg1, dArg2);
    break;
  case tcXy:
    dRes = pow(dArg1, dArg2);
    break;
  case tcYvx:
    dRes = pow(dArg1, 1.0/dArg2);
    break;
  default:
    iRet = -1;
   }
  return iRet;
}

// *******************************************************************************
char cKeyToChar[17] =
{
  '0', // 0
  '1', // 1
  '2', // 2
  '3', // 3
  '4', // 4
  '5', // 5
  '6', // 6
  '7', // 7
  '8', // 8
  '9', // 9
  'a', // a
  'b', // b
  'c', // c
  'd', // d
  'e', // e
  'f', // f
  'r', // r
};

class CDisp
{
  int               m_iTab;
  int               m_iMaxTab;
  int               m_iBase;
  int               m_i2nd;
  int               m_bFix;
  char              m_strDispFull[33];
  char              m_strDispTab[16];

  CStackDoubleExp   m_dStack;
  CStackIntExp      m_iStack;
  int               m_iMode;
  double            m_dAccu;
  double            m_dMem;
  int               m_bDAccuV;
  int               m_iE;
  int               m_iFix;
  int               m_bErr;

  int               m_bIAccuV;
  int               m_iAccu;

  int               DispLen(char * lpStr);
  void              InvKey(void){g_print("\a");}
  void              Error(void);
  int               StrToDouble(void);
  int               DoubleToStr(void);

  void              EvConst(KeyTokens tcKey);
  void              EvNumber(KeyTokens tcKey);
  void              EvDel(KeyTokens tcKey);
  void              EvEtc(KeyTokens tcKey);

  void              EvTrig(KeyTokens tcKey);
  int               EvParen(KeyTokens tcKey);
  void              EvOp2(KeyTokens tcOp);
  void              EvMem(KeyTokens tcKey);


  void              EvOp1(KeyTokens tcOp);
  int               EvBase(KeyTokens tcOp);
  void              EvSize(void){}
  int               IntToStr(void);
  int               StrToInt(void);
public:
  CDisp();
  void              KeyEvent(int iKey);
  int               KeyToCode(int iKey);
  char            * GetDisp(int &iBase, int &iMode, int &iMaxTab, int &iTab, int &i2nd, int &iFix, int &iE);
};

int CDisp::DispLen(char * lpStr)
{
  int i = 0;
  while(*lpStr != '\0')
  {
    if(*lpStr != '-' && *lpStr != '.')
      i++;
    lpStr++;
  }
  return i;
}

void CDisp::Error(void)
{
  if(m_iBase == 1)
  {
    m_dStack.Delete();
    m_dAccu   = 0;
    m_bDAccuV = 1;
  }
  else
  {
    m_iStack.Delete();
    m_iAccu   = 0;
    m_bIAccuV = 1;
  }
  m_bErr    = 0;
  InvKey();
}

int CDisp::StrToDouble(void)
{
  if(!m_bDAccuV)
  {
    char * lpEnd;
    m_bDAccuV = 1;
    m_dAccu = strtod(m_strDispTab, &lpEnd);
    if(*lpEnd != '\0' || errno != 0)
      return -1;
  }
  return 0;
}

int CDisp::StrToInt(void)
{
  int iBase = !m_iBase ? 16 : m_iBase == 2 ? 8 : 2;
  if(!m_bIAccuV)
  {
    char * lpEnd;
    m_bIAccuV = 1;
    m_iAccu = (int) strtoul(m_strDispFull, &lpEnd, iBase);
    if(*lpEnd != '\0' || errno != 0)
      return -1;
  }
  return 0;
}

int CDisp::IntToStr(void)
{
  if(m_bIAccuV)
  {
    switch(m_iBase)
    {
    case 0:
      sprintf(m_strDispFull, "%x", m_iAccu);
      break;
    case 2:
      sprintf(m_strDispFull, "%o", m_iAccu);
      break;
    case 3:
      {
        int b = 0;
        int i = 32, j = m_iAccu;
        char * cp = m_strDispFull;
        while(i--)
        {
          if(j & 0x80000000)
          {
            b = 1;
            *cp++ = '1';
          }
          else if(b)
            *cp++ = '0';
          j *= 2;
        }
        if(cp == m_strDispFull)
          *cp++ = '0';
        *cp = '\0';
      }
      break;
    default:
      return -1;
    }
  }
  return 0;
}

int CDisp::DoubleToStr(void)
{
  char *cpSrc, *cpDst = m_strDispFull;
  int  i, iDec, iSign, iExp = (m_dAccu == 0) ? 0 : (int)log10(fabs(m_dAccu));

  if(!m_bDAccuV)
    return 0;

  *m_strDispFull = '\0';
  if(iExp >= +100.0) return -1;
  if(iExp <= -100.0) m_dAccu = 0;

  if(!m_iE && m_iFix >= 0 && m_iFix <= 11)
  {
    if(!m_iE && iExp < 12 - m_iFix && iExp >= 0)
    {
      i = 11 - iExp >= m_iFix ? m_iFix : 11 - iExp;
      sprintf(m_strDispFull, "%#.*f", i, m_dAccu);
      goto end;
    }
    if(!m_iE && iExp < 0 && iExp + 11 >= 0)
    {
      sprintf(m_strDispFull, "%#.*f", m_iFix, m_dAccu);
      goto end;
    }
    i = m_iFix >= 8 ? 9 : m_iFix + 1;
    cpSrc = ecvt(m_dAccu, i, &iDec, &iSign);
    i = 1;
    iDec--;
    if(iSign)
      *cpDst++ = '-';
    do
    {
      if(!i--)*cpDst++ = '.';
      *cpDst++ = *cpSrc++;
    }
    while(*cpSrc != '\0');
    sprintf(cpDst, "e%02i", iDec);
    goto end;
  }
  if(iExp <= 11 && iExp > -5 && !m_iE)
  {
    sprintf(m_strDispFull, "%#.*f", iExp >= 0 ? 11 - iExp : 11, m_dAccu);
    for(i = strlen(m_strDispFull) - 1; i > 0; i--)
      if(m_strDispFull[i] == '0')
         m_strDispFull[i] = '\0';
      else
         break;
    goto end;
  }
  cpSrc = ecvt(m_dAccu, 9, &iDec, &iSign);
  if(m_iE == 1)
  {
    if(m_dAccu != 0)
    {
      if(iDec > 0)
      {
        i = (iDec - 1) % 3;
        i += 1;
        iDec -= i;
      }
      else if(iDec == 0)
      {
        i = iDec % 3;
        i += 3;
        iDec -= i;
      }
      else
      {
        i = (-iDec) % 3;
        i = 3 - i;
        iDec -= i;
      }
    }
    else
      i = iDec--;
  }
  else
  {
    i = 1;
    iDec--;
  }
  if(iSign)
    *cpDst++ = '-';
  do
  {
    if(!i--)*cpDst++ = '.';
    *cpDst++ = *cpSrc++;
  }
  while(*cpSrc != '\0');
  while(*(--cpDst) == '0');
  sprintf(++cpDst, "e%02i", iDec);
  goto end;
end:
//  g_print("Exp: %i Num: %.12f m_strDispFull: %s\n", iExp, m_dAccu, m_strDispFull);
  if(strlen(m_strDispFull) > 15)
  {
    g_print("String too long: - StrToDouble\n");
    exit(-1);
  }
  strcpy(m_strDispTab, m_strDispFull);
  return 0;
}

CDisp::CDisp()
{
  *m_strDispFull  = '\0';
  *m_strDispTab   = '\0';
  m_iTab          =  0;
  m_iMaxTab       =  0;
  m_iBase         =  1;
  m_i2nd          =  0;
  m_bFix          =  0;

  m_iMode         =  0;
  m_dAccu         =  0;
  m_dMem          =  0;
  m_bDAccuV       =  1;
  m_iE            =  0;
  m_iFix          =  -1;
  m_bErr          =  0;

  m_bIAccuV       = 1;
  m_iAccu         = 0;
}

void CDisp::EvNumber(KeyTokens tcKey)
{
  char strTmp[2] = {0,0};
  char *cp;
  int iDispLen, iMin, iMax, iNum = (int)(tcKey - tc0);
  switch(m_iBase)
  {
  case 0:
    iMin = 0;
    iMax = 15;
    iDispLen = 8;
    break;
  case 1:
    iMin = 0;
    iMax = 9;
    iDispLen = 12;
    break;
  case 2:
    iMin = 0;
    iMax = 7;
    iDispLen = 11;
    break;
  case 3:
    iMin = 0;
    iMax = 1;
    iDispLen = 32;
    break;
  default:
    g_print("Invalid key: - EcNumber\n");
    exit(-1);
  }
  if(iNum < iMin || iNum > iMax)
  {
    InvKey();
    goto end;
  }
  if(m_iBase == 1)
  {
    if(m_bDAccuV)
      *m_strDispTab = '\0';
    if((cp = strchr(m_strDispTab, 'e')) || (cp = strchr(m_strDispTab, 'E')))
    {
      if(*(++cp) == '-')
        cp++;
      if(strlen(cp) >= 2)
      {
        InvKey();
        goto end;
      }
    }
    if(DispLen(m_strDispTab) < iDispLen)
    {
      m_bDAccuV = 0;
      *strTmp = cKeyToChar[iNum];
      strcat(m_strDispTab, strTmp);
    }
    else
      InvKey();
  }
  else
  {
    if(m_bIAccuV)
      *m_strDispFull = '\0';
    if(strlen(m_strDispFull) < (size_t)iDispLen)
    {
      m_bIAccuV = 0;
      *strTmp = cKeyToChar[iNum];
      strcat(m_strDispFull, strTmp);
    }
    else
      InvKey();
  }
end:
  return;
}

void CDisp::EvDel(KeyTokens tcKey)
{
  int iLen;
  if(m_iBase == 1)
  {
    switch(tcKey)
    {
    case tcBsp:
      iLen = strlen(m_strDispTab);
      if(iLen > 1)
      {
        m_strDispTab[iLen - 1] = '\0';
        m_bDAccuV = 0;
        break;
      }
    case tcCe:
      m_bDAccuV = 1;
      m_dAccu = 0;
      break;
    case tcAc:
      m_dStack.Delete();
      m_bDAccuV = 1;
      m_dAccu = 0;
      m_iStack.Delete();
      m_bIAccuV = 1;
      m_iAccu = 0;
      break;
    default:
      g_print("Invalid key: - EcDel\n");
      exit(-1);
    }
  }
  else
  {
    switch(tcKey)
    {
    case tcBsp:
      iLen = strlen(m_strDispFull);
      if(iLen > 1)
      {
        m_strDispFull[iLen - 1] = '\0';
        m_bIAccuV = 0;
        break;
      }
    case tcCe:
      m_bIAccuV = 1;
      m_iAccu = 0;
      break;
    case tcAc:
      m_dStack.Delete();
      m_bDAccuV = 1;
      m_dAccu = 0;
      m_iStack.Delete();
      m_bIAccuV = 1;
      m_iAccu = 0;
      break;
    default:
      g_print("Invalid key: - EcDel1\n");
      exit(-1);
    }
  }
}

void CDisp::EvMem(KeyTokens tcKey)
{
  if(m_iBase == 1)
  {
    switch(tcKey)
    {
    case tcM:
      m_dMem = m_dAccu;
      break;
    case tcMp:
      m_dMem += m_dAccu;
      break;
    case tcMr:
      m_dAccu = m_dMem;
      m_bDAccuV = 1;
      break;
    case tcMc:
      m_dMem = 0.0;
      break;
    default:
      g_print("Invalid key: - EvMem\n");
      exit(-1);
    }
  }
  else
    switch(tcKey)
    {
    case tcM:
      m_dMem = m_iAccu;
      break;
    case tcMp:
      m_dMem += m_iAccu;
      break;
    case tcMr:
      if(m_dMem <= INT_MAX && m_dMem >= INT_MIN)
      {
        m_iAccu = (int)m_dMem;
        m_bIAccuV = 1;
      }
      else
        InvKey();
      break;
    case tcMc:
      m_dMem = 0.0;
      break;
    default:
      g_print("Invalid key: - EvMem2\n");
      exit(-1);
    }
}

void CDisp::EvTrig(KeyTokens tcKey)
{
  static double d2Pi = 2 * M_PI;
  double dMode = d2Pi;
  if(m_iBase != 1)
  {
    InvKey();
    return;
  }
  if(m_iMode)
    dMode = 400.0;
  else
    dMode = 360.0;
  if(tcKey <= tcTan && (m_iMode == 0 || m_iMode == 2))
    m_dAccu /= dMode / d2Pi;
  switch(tcKey)
  {
  case tcSin:
    m_dAccu = sin(m_dAccu);
    break;
  case tcCos:
    m_dAccu = cos(m_dAccu);
    break;
  case tcTan:
    m_dAccu = tan(m_dAccu);
    break;
  case tcAsin:
    m_dAccu = asin(m_dAccu);
    break;
  case tcAcos:
    m_dAccu = acos(m_dAccu);
    break;
  case tcAtan:
    m_dAccu = atan(m_dAccu);
    break;
  default:
    g_print("Invalid key: - EvTrig\n");
    exit(-1);
  }
  if(tcKey > tcTan && (m_iMode == 0 || m_iMode == 2))
    m_dAccu *= dMode / d2Pi;
  m_bDAccuV = 1;
}

void CDisp::EvEtc(KeyTokens tcKey)
{
  char * cp;
  if(m_iBase != 1)
  {
    InvKey();
    goto end;
  }
  switch(tcKey)
  {
  case tcMin:
    if(strlen(m_strDispTab) == 0)
    {
      InvKey();
      goto end;
    }
    if(*m_strDispTab == '-')
      memmove(m_strDispTab, m_strDispTab + 1, strlen(m_strDispTab));
    else
    {
      memmove(m_strDispTab + 1, m_strDispTab, strlen(m_strDispTab) + 1);
      *m_strDispTab = '-';
    }
    break;
  case tcDot:
    if(m_bDAccuV)
      *m_strDispTab = '\0';
    if(strchr(m_strDispTab, '.') || strchr(m_strDispTab, 'e') || strchr(m_strDispTab, 'E'))
    {
      InvKey();
      goto end;
    }
    else
      strcat(m_strDispTab, ".");
    break;
  case tcExp:
    if(strlen(m_strDispTab) == 0)
    {
      InvKey();
      goto end;
    }
    if(strchr(m_strDispTab, 'e') || strchr(m_strDispTab, 'E'))
    {
      InvKey();
      goto end;
    }
    else
      strcat(m_strDispTab, "e");
    break;
  case tcEmin:
    if(!((cp = strchr(m_strDispTab, 'e')) || (cp = strchr(m_strDispTab, 'E'))))
    {
      InvKey();
      goto end;
    }
    if(*(++cp) == '\0')
    {
      InvKey();
      goto end;
    }
    if(*cp == '-')
      memmove(cp, cp + 1, strlen(cp));
    else
    {
      memmove(cp + 1, cp, strlen(cp) + 1);
      *cp = '-';
    }
    break;
  default:
    g_print("Invalid key: - EvDel\n");
    exit(-1);
  }
  m_bDAccuV = 0;
end:
  return;
}

int CDisp::EvParen(KeyTokens tcKey)
{
  int iRet = 0;
  if(tcKey == tcPl)
  {
    if(m_dStack.AddLP(m_dAccu) < 0)
      iRet = -1;
  }
  else
  {
    if(m_dStack.AddRP(m_dAccu) < 0)
      iRet = 0;
  }
  return iRet;
}

void CDisp::EvOp1(KeyTokens tcOp)
{
  switch(m_iBase)
  {
  case 1:
    switch(tcOp)
    {
    case tcEq:
      if(m_dStack.AddOp(m_dAccu, m_dAccu) < 0)
      {
        Error();
        break;
      }
      if(m_dStack.ExeAll(m_dAccu) < 0)
      {
        Error();
        break;
      }
      break;
    case tcX1:
      m_dAccu = pow(m_dAccu, -1);
      break;
    case tcX2:
      m_dAccu = pow(m_dAccu, 2);
      break;
    case tcX3:
      m_dAccu = pow(m_dAccu, 3);
      break;
    case tc2vx:
      m_dAccu = pow(m_dAccu, 1.0/2.0);
      break;
    case tc3vx:
      m_dAccu = pow(m_dAccu, 1.0/3.0);
      break;
    case tcLn:
      m_dAccu = log(m_dAccu);
      break;
    case tcEx:
      m_dAccu = exp(m_dAccu);
      break;
    case tcLog:
      m_dAccu = log10(m_dAccu);
      break;
    case tc10x:
      m_dAccu = pow(10.0, m_dAccu);
      break;
    default:
      g_print("InvToken - EvOp1\n");
    }
    m_bDAccuV = 1;
    break;
  default:
    switch(tcOp)
    {
    case tcEq:
      if(m_iStack.AddOp(m_iAccu, m_iAccu) < 0)
      {
        Error();
        break;
      }
      if(m_iStack.ExeAll(m_iAccu) < 0)
      {
        Error();
        break;
      }
      break;
    case tcNeg:
      m_iAccu = -m_iAccu;
      break;
    default:
      g_print("InvToken - EvOp1 \n");
    }
    m_bDAccuV = 1;
    break;
  }
}


void CDisp::EvOp2(KeyTokens tcOp)
{
  switch(m_iBase)
  {
  case 1:
    if(m_dStack.AddOp(m_dAccu, m_dAccu) < 0)
    {
      Error();
      break;
    }
    if(m_dStack.AddOp(tcOp, m_dAccu) < 0)
    {
      Error();
      break;
    }
    m_bDAccuV = 1;
    break;
  default:
    if(m_iStack.AddOp(m_iAccu, m_iAccu) < 0)
    {
      Error();
      break;
    }
    if(m_iStack.AddOp(tcOp, m_iAccu) < 0)
    {
      Error();
      break;
    }
    m_bIAccuV = 1;
  }
}

char * CDisp::GetDisp(int &iBase, int &iMode, int &iMaxTab, int &iTab, int &i2nd, int &iFix, int &iE)
{
  m_iMaxTab = 0;
  if(m_iBase == 1)
  {
    if(m_bDAccuV)
    {
      if(DoubleToStr())
        m_bErr = 1;
    }
    if(m_bErr)
    {
      strcpy(m_strDispTab, "Err");
      Error();
    }
  }
  else
  {
    if(m_bIAccuV)
    {
      if(IntToStr())
        m_bErr = 1;
    }
    switch(m_iBase)
    {
    case 0:
      if(m_bErr)
      {
        strcpy(m_strDispTab, "Err");
        Error();
      }
      else
      {
        strncpy(m_strDispTab, m_strDispFull, 8);
        m_strDispTab[8] = '\0';
      }
      break;
    case 2:
      if(m_bErr)
      {
        strcpy(m_strDispTab, "Err");
        Error();
      }
      else
      {
        strncpy(m_strDispTab, m_strDispFull, 11);
        m_strDispTab[11] = '\0';
      }
      break;
    case 3:
      if(m_bErr)
      {
        strcpy(m_strDispTab, "Err");
        Error();
      }
      else
      {
        int iOff, iLen = strlen(m_strDispFull);
        m_iMaxTab = (iLen - 1) / 8 + 1;
        m_iTab = m_iTab % m_iMaxTab;

        if((iOff = iLen - (m_iTab + 1) * 8) < 0)
          iOff = 0;
        if((iLen -= m_iTab * 8) > 8)
          iLen = 8;
        strncpy(m_strDispTab, m_strDispFull + iOff, iLen);
        m_strDispTab[iLen] = '\0';
      }
      break;
    default:
      ;
    }
  }
  iBase   = m_iBase;
  iMode   = m_iMode;
  iMaxTab = m_iMaxTab;
  iTab    = m_iTab;
  i2nd    = m_i2nd;
  iFix    = m_iFix;
  iE      = m_iE;
  return m_strDispTab;
}

void CDisp::EvConst(KeyTokens tcKey)
{
  switch(tcKey)
  {
  case tcPi:
    m_dAccu = M_PI;
    break;
  case tcEe:
    m_dAccu = M_E;
    break;
  default:
    g_print("Invalid key: - EcConst\n");
    exit(-1);
  }
  m_bDAccuV = 1;
}

int CDisp::EvBase(KeyTokens tcOp)
{
  int iRet = 0;
  int iBase = (int)(tcOp - tcHex);
  if(iBase == m_iBase)
    goto End;
  if(m_iBase == 1)
  {
    if(StrToDouble())
    {
      iRet = -1;
      goto End;
    }
    if(m_dAccu > INT_MAX || m_dAccu < INT_MIN)
    {
      InvKey();
      goto End;
    }
    m_iAccu = (int)m_dAccu;
    m_bIAccuV = 1;
    m_iBase = iBase;
  }
  else
  {
    if(StrToInt())
    {
      iRet = -1;
      goto End;
    }
    m_dAccu = m_iAccu;
    m_bDAccuV = 1;
    m_iBase = iBase;
  }
End:
  return iRet;
}


void CDisp::KeyEvent(int iKey)
{
  KeyTokens tcKey;
  errno = 0;
  m_bErr = 0;
  if(m_iBase == 1)
  {
    if(!m_i2nd)
      tcKey = tableNorm[iKey];
    else
      tcKey = table2nd[iKey];
  }
  else
  {
    if(!m_i2nd)
      tcKey = tableBaseNNorm[iKey];
    else
      tcKey = tableBaseN2nd[iKey];
  }
  if(m_bFix)
  {
    if(tcKey >= tc0 && tcKey <= tc6)
      m_iFix = (int)(tcKey - tc0);
    else
      m_iFix = -1;
    m_bFix = 0;
    if(tcKey < tc0 || tcKey >tc9)
      InvKey();
    goto end;
  }
  if(tcKey == tc2nd)
  {
    m_i2nd = !m_i2nd;
    goto end;
  }
  m_i2nd = 0;
// no need conversion
  if(tcKey >= tcPi && tcKey <=tcEe)
  {
    EvConst(tcKey);
    goto end;
  }
  if(tcKey >= tcDeg && tcKey <= tcGrad)
  {
    m_iMode = (int)(tcKey - tcDeg);
    goto end;
  }
  if(tcKey == tcTab && m_iBase == 3)
  {
    m_iTab++;
    goto end;
  }
  if(m_iBase == 1)
  {
//double to display
    if(tcKey >= tc0 && tcKey <= tcDot)
    {
      if((m_bErr = DoubleToStr()))
        goto end;
      if(tcKey >= tc0 && tcKey <= tc9)
      {
        EvNumber(tcKey);
        goto end;
      }
      if(tcKey >= tcBsp && tcKey <= tcAc)
      {
        EvDel(tcKey);
        goto end;
      }
      if(tcKey >= tcEmin && tcKey <= tcDot)
      {
        EvEtc(tcKey);
        goto end;
      }
    }
  }
  else
  {
    if(tcKey >= tc0 && tcKey <= tcAc)
    {
      if((m_bErr = IntToStr()))
        goto end;
      if(tcKey >= tc0 && tcKey <= tcF)
      {
        EvNumber(tcKey);
        goto end;
      }
      if(tcKey >= tcBsp && tcKey <= tcAc)
      {
        EvDel(tcKey);
        goto end;
      }
    }
  }
  if(tcKey >= tcHex && tcKey <= tcBin)
  {
    m_bErr = EvBase(tcKey);
    goto end;
  }
// string to double
  if(m_iBase == 1)
  {
    if(tcKey >= tcXy && tcKey <= tcMc)
    {
      if((m_bErr = StrToDouble()))
        goto end;
      if(tcKey >= tcSin && tcKey <= tcAtan)
      {
        EvTrig(tcKey);
        goto end;
      }
      if(tcKey >= tcPl && tcKey <= tcPr)
      {
        m_bErr = EvParen(tcKey);
        goto end;
      }
      if(tcKey >= tcLn && tcKey <= tcEq)
      {
        EvOp1(tcKey);
        goto end;
      }
      if(tcKey >= tcXy && tcKey <= tcMod)
      {
        EvOp2(tcKey);
        goto end;
      }
      if(tcKey >= tcM && tcKey <= tcMc)
      {
        EvMem(tcKey);
        goto end;
      }
      if(tcKey == tcFix && !m_iE)
      {
        m_bFix = 1;
        goto end;
      }
      if(tcKey == tcFe && m_iBase == 1)
      {
        m_iE = (m_iE + 1) % 3;
        goto end;
      }
    }
  }
  else
  {
    if(tcKey >= tcMul && tcKey <= tcMod || tcKey >= tcM && tcKey <= tcNeg || tcKey == tcEq)
    {
      if((m_bErr = StrToInt()))
        goto end;
      if(tcKey >= tcMul && tcKey <= tcMod || tcKey >= tcAnd && tcKey <= tcXor)
      {
        EvOp2(tcKey);
        goto end;
      }
      if(tcKey >= tcM && tcKey <= tcMc)
      {
        EvMem(tcKey);
        goto end;
      }
      if(tcKey == tcEq || tcKey == tcNeg)
      {
        EvOp1(tcKey);
        goto end;
      }
    }
  }
  InvKey();
end:
  if(errno)
    m_bErr = -1;
  return;
}

static int iNums[] =
{
  38, 19, 20, 21, 25, 26, 27, 31, 32, 33
};

int CDisp::KeyToCode(int iKey)
{
  int iRet = -1;
  if(iKey >= 48 && iKey <= 57 || iKey >= 65456 && iKey <= 65465)
  {
    iRet = iNums[iKey > 57 ? iKey - 65456 : iKey - 48];
    goto End;
  }
  if(!m_iBase && (iKey >= 97 && iKey <= 102 || iKey >= 65 && iKey <= 70))
  {
    iRet = (iKey > 70 ? iKey - 97 : iKey - 65) + 6;
    goto End;
  }
  if(m_iBase == 1 && iKey == 32)
  {
    iRet = 0;
    goto End;
  }
  if(iKey == 65439 || iKey == 65535)
  {
    iRet = 23;
    goto End;
  }
  if(iKey == 65455 || iKey == 47)
  {
    iRet = 22;
    goto End;
  }
  if(iKey == 65450 || iKey == 42)
  {
    iRet = 28;
    goto End;
  }
  if(iKey == 65453 || iKey == 45)
  {
    iRet = 34;
    goto End;
  }
  if(iKey == 65451 || iKey == 43)
  {
    iRet = 40;
    goto End;
  }
  if(iKey == 65421 || iKey == 65293)
  {
    iRet = 41;
    goto End;
  }
  if(iKey == 65454 || iKey == 46)
  {
    iRet = 39;
    goto End;
  }
  if(iKey == 65288)
  {
    iRet = 17;
    goto End;
  }
  if(iKey == 37)
  {
    iRet = 35;
    goto End;
  }
  if(iKey >= 65505 && iKey<= 65508)
    goto End;
  InvKey();
End:
  return iRet;
}

CDisp disp;

char * GetDisp(int &iBase, int &iMode, int &iMaxTab, int &iTab, int &i2nd, int &iFix, int &iE)
{
  char * cp = disp.GetDisp(iBase, iMode, iMaxTab, iTab, i2nd, iFix, iE);
  return cp;
}

int KeyEvent(int iKey, int iMode)
{
//  g_print("Key: %i\n", iKey);
  if(iMode)
  {
    int i = disp.KeyToCode(iKey);
    if(i >= 0) disp.KeyEvent(i);
    else return -1;
  }
  else
    disp.KeyEvent(iKey);
  return 0;
}
